
#include <sys/wait.h>

#include "localserver.h"


#define MAX_CONNECTIONS 5


#define MAX_RUNNING_THREADS 10
static pid_t running_pids[MAX_RUNNING_THREADS]; 

static int nr_running_threads = 0;

static void InitRunningThreads()
{
  nr_running_threads = 0;
  int i;
  for (i=0;i<MAX_RUNNING_THREADS;i++)
  {
    running_pids[i] = -1;
  }
}

static void AddThreadPid(pid_t newPid)
{
  int i;
  for (i=0;i<MAX_RUNNING_THREADS;i++)
  {
    if (running_pids[i] == -1)
    {
      running_pids[i] = newPid;
      nr_running_threads++;
      break;
    }
  }
}

static void CheckForZombieProcesses()
{
  int i;
  for (i=0;i<MAX_RUNNING_THREADS;i++)
  {
    if (running_pids[i] != -1)
    {
      int status;
      if (wait4(running_pids[i],&status,WNOHANG|WUNTRACED,NULL) == running_pids[i])
      {
        //printf("Found zombie process with id %d\n",running_pids[i]);
        // External application crashed? Bring TTGO to the foreground
        running_pids[i] = -1;
        nr_running_threads--;
      }
    }
  }
}


int localserver(char* the_address,void (*requesthandler)(char*,int))
{
    InitRunningThreads();
    int server_sockfd, client_sockfd;
    socklen_t server_len, client_len;
    struct sockaddr_un server_address;
    struct sockaddr_un client_address;
    int result;
    fd_set readfds, testfds;
  

    int serverbusy;
    serverbusy=1;

    signal(SIGPIPE,SIG_IGN);
    
    unlink(the_address);

    server_sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
    server_address.sun_family = AF_UNIX;
    strcpy(server_address.sun_path, the_address);
    server_len = sizeof(server_address);

    if (bind(server_sockfd, (struct sockaddr *)&server_address, server_len))
    {
        perror("Server Could not bind to the socket\n");
        exit(1);
    }
    if (listen(server_sockfd, MAX_CONNECTIONS))
    {
        perror("Server Could not listen to the socket\n");
        exit(1);
    }

    FD_ZERO(&readfds);
    FD_SET(server_sockfd, &readfds);

    while(serverbusy)
    {
        int fd;
        int nread;
        testfds = readfds;

        CheckForZombieProcesses();
        if (nr_running_threads == MAX_RUNNING_THREADS)
          continue;

        //LOGTODO        printf("server waiting\n");
        result = select(FD_SETSIZE, &testfds, (fd_set *)0, (fd_set *)0, (struct timeval *) 0);

        if(result < 1)
        {
            perror("server5");
            exit(1);
        }
        for(fd = 0; fd < FD_SETSIZE; fd++)
        {
            if(FD_ISSET(fd,&testfds))
            {
                if(fd == server_sockfd)
                {
                    client_len = sizeof(client_address);
                    client_sockfd = accept(server_sockfd, 
                        (struct sockaddr *)&client_address, (MY_SL_TYPE*)&client_len);
#ifdef __CYGWIN__
                    if (client_sockfd != 0xffffffff)
#endif
                    {
                        FD_SET(client_sockfd, &readfds);
                        //LOGTODO                    printf("adding client on fd %d\n", client_sockfd);
                    }
                }
                else
                {
                    ioctl(fd, FIONREAD, &nread);

                    if(nread == 0)
                    {
                        close(fd);
                        FD_CLR(fd, &readfds);
                        //LOGTODO                        printf("removing client on fd %d\n", fd);
                    }
                    else 
                    {
                      pid_t childpid = fork();
                      if (childpid == 0) 
                      {
                        //TODO PLEASECHECK optimize!
                          char *request = (char *)malloc(nread+1);
                          if (request)
                          {
                              char* buffer = request;
                              read(fd, buffer, nread);
                              buffer[nread]='\0';
                  			      requesthandler(buffer,fd);
                              free(request);
                          }
                          else
                          {
                              //TODO show an error not enough memory!
                          }
                        
                          close(fd);
                          FD_CLR(fd, &readfds);
                          //LOGTODO                        printf("removing client on fd %d\n", fd);
                          exit(0);
                      }
                      else
                      {
                        AddThreadPid(childpid);
                        close(fd);
                        FD_CLR(fd, &readfds);
                      }                 
                    }
                }
            }
        }
    }
}



